import warnings
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Any, Callable, Generator, Optional, Union

from attrs import evolve, frozen

from ..types import ExperimentalFeatureWarning


@frozen
class Scope:
    record_metric: Callable[[str, Union[float, int]], None]
    _tag: Optional[str] = None


_current_scope: ContextVar[Optional[Scope]] = ContextVar("scope", default=None)


def current_scope() -> Scope:
    warnings.warn(
        "current_scope is an experimental internal function. It may change or be removed without warning.",
        category=ExperimentalFeatureWarning,
        stacklevel=1,
    )
    return _get_current_scope()


def _get_current_scope() -> Scope:
    s = _current_scope.get()
    if s is None:
        raise RuntimeError("No scope available")
    return s


@contextmanager
def scope(sc: Scope) -> Generator[None, None, None]:
    s = _current_scope.set(sc)
    try:
        yield
    finally:
        _current_scope.reset(s)


@contextmanager
def evolve_scope(**kwargs: Any) -> Generator[None, None, None]:
    new_scope = evolve(_get_current_scope(), **kwargs)
    s = _current_scope.set(new_scope)
    try:
        yield
    finally:
        _current_scope.reset(s)



import os
import sys
import uuid

import pytest

from cog.code_xforms import load_module_from_string, strip_model_source_code

g_module_dir = os.path.dirname(os.path.abspath(__file__))


@pytest.mark.skipif(sys.version_info < (3, 9), reason="requires python3.9 or higher")
def test_train_function_model():
    with open(f"{g_module_dir}/server/fixtures/train.py", encoding="utf-8") as file:
        source_code = file.read()

    new_source = strip_model_source_code(source_code, "train", "train")
    expected_source = """
from cog import BaseModel, Input, Path
class TrainingOutput(BaseModel):
    weights: Path
def train(n: int=Input(description='Dimension of weights to generate')) -> TrainingOutput:
    return None
"""
    assert expected_source.strip() == new_source.strip()
    assert load_module_from_string(uuid.uuid4().hex, new_source)


@pytest.mark.skipif(sys.version_info < (3, 9), reason="requires python3.9 or higher")
def test_predict_many_inputs():
    with open(
        f"{g_module_dir}/../../test-integration/test_integration/fixtures/many-inputs-project/predict.py",
        encoding="utf-8",
    ) as file:
        source_code = file.read()

    new_source = strip_model_source_code(source_code, ["Predictor"], ["predict"])
    expected_source = """
from cog import BasePredictor, Input, Path
class Predictor(BasePredictor):

    def predict(self, no_default: str, default_without_input: str='default', input_with_default: int=Input(default=10), path: Path=Input(description='Some path'), image: Path=Input(description='Some path'), choices: str=Input(choices=['foo', 'bar']), int_choices: int=Input(description='hello', choices=[3, 4, 5])) -> str:
        return None
"""
    assert expected_source.strip() == new_source.strip()
    assert load_module_from_string(uuid.uuid4().hex, new_source)


@pytest.mark.skipif(sys.version_info < (3, 9), reason="requires python3.9 or higher")
def test_predict_output_path_model():
    with open(
        f"{g_module_dir}/../../test-integration/test_integration/fixtures/path-output-project/predict.py",
        encoding="utf-8",
    ) as file:
        source_code = file.read()

    new_source = strip_model_source_code(source_code, ["Predictor"], ["predict"])
    expected_source = """
import os
from cog import BasePredictor, Path
class Predictor(BasePredictor):

    def predict(self) -> Path:
        return None
"""
    assert expected_source.strip() == new_source.strip()
    assert load_module_from_string(uuid.uuid4().hex, new_source)


@pytest.mark.skipif(sys.version_info < (3, 9), reason="Requires Python 3.9 or newer")
def test_strip_model_source_code():
    stripped_code = strip_model_source_code(
        """
import io

from cog import BasePredictor, Path
from typing import Optional
from pydantic import BaseModel
import torch


class ModelOutput(BaseModel):
    success: bool
    error: Optional[str]
    segmentedImage: Optional[Path]


class Predictor(BasePredictor):
    # setup code
    def predict(self, msg: str) -> ModelOutput:
       return ModelOutput(success=False, error=msg, segmentedImage=None)
""",
        ["Predictor"],
        ["predict"],
    )
    assert (
        stripped_code
        == """from cog import BasePredictor, Path
from typing import Optional
from pydantic import BaseModel
class ModelOutput(BaseModel):
    success: bool
    error: Optional[str]
    segmentedImage: Optional[Path]
class Predictor(BasePredictor):

    def predict(self, msg: str) -> ModelOutput:
        return None"""
    ), "Stripped code needs to equal the minimum viable type inference."


@pytest.mark.skipif(sys.version_info < (3, 9), reason="Requires Python 3.9 or newer")
def test_strip_model_source_code_removes_function_decorators():
    stripped_code = strip_model_source_code(
        """
import io

from cog import BasePredictor, Path
from typing import Optional
from pydantic import BaseModel
import torch


class ModelOutput(BaseModel):
    success: bool
    error: Optional[str]
    segmentedImage: Optional[Path]


class Predictor(BasePredictor):

    # setup code
    @torch.inference_mode()
    def predict(self, msg: str) -> ModelOutput:
       return ModelOutput(success=False, error=msg, segmentedImage=None)
""",
        ["Predictor"],
        ["predict"],
    )
    assert (
        stripped_code
        == """from cog import BasePredictor, Path
from typing import Optional
from pydantic import BaseModel
class ModelOutput(BaseModel):
    success: bool
    error: Optional[str]
    segmentedImage: Optional[Path]
class Predictor(BasePredictor):

    def predict(self, msg: str) -> ModelOutput:
        return None"""
    ), "Stripped code needs to equal the minimum viable type inference."


@pytest.mark.skipif(sys.version_info < (3, 9), reason="Requires Python 3.9 or newer")
def test_strip_model_source_code_keeps_referenced_globals():
    stripped_code = strip_model_source_code(
        """
import io

from cog import BasePredictor, Path
from typing import Optional
from pydantic import BaseModel
import torch
import numpy as np


INPUT_DIMS = list(np.arange(32, 64, 32))


class ModelOutput(BaseModel):
    success: bool
    error: Optional[str]
    segmentedImage: Optional[Path]


class Predictor(BasePredictor):
    # setup code
    def predict(self, height: int=Input(description='Height of image', default=128, choices=INPUT_DIMS)) -> ModelOutput:
       return ModelOutput(success=False, error=msg, segmentedImage=None)
""",
        ["Predictor"],
        ["predict"],
    )
    assert (
        stripped_code
        == """from cog import BasePredictor, Path
from typing import Optional
from pydantic import BaseModel
import numpy as np
INPUT_DIMS = list(np.arange(32, 64, 32))
class ModelOutput(BaseModel):
    success: bool
    error: Optional[str]
    segmentedImage: Optional[Path]
class Predictor(BasePredictor):

    def predict(self, height: int=Input(description='Height of image', default=128, choices=INPUT_DIMS)) -> ModelOutput:
        return None"""
    ), "Stripped code needs to equal the minimum viable type inference."


@pytest.mark.skipif(sys.version_info < (3, 9), reason="Requires Python 3.9 or newer")
def test_strip_model_source_code_keeps_referenced_subclasses():
    stripped_code = strip_model_source_code(
        """
import io

from cog import BasePredictor, Path
from typing import Optional
from pydantic import BaseModel
import torch
import numpy as np


INPUT_DIMS = list(np.arange(32, 64, 32))


class ModelOutput(BaseModel):
    success: bool
    error: Optional[str]
    segmentedImage: Optional[Path]


class Predictor(BasePredictor):
    # setup code
    def predict(self, height: int=Input(description='Height of image', default=128, choices=INPUT_DIMS)) -> ModelOutput:
       return ModelOutput(success=False, error=msg, segmentedImage=None)

class SchnellPredictor(Predictor):
    # setup code
    def predict(self, height: int=Input(description='Height of image', default=128, choices=INPUT_DIMS)) -> ModelOutput:
       return ModelOutput(success=False, error=msg, segmentedImage=None)
""",
        ["SchnellPredictor"],
        ["predict"],
    )
    assert (
        stripped_code
        == """from cog import BasePredictor, Path
from typing import Optional
from pydantic import BaseModel
import numpy as np
INPUT_DIMS = list(np.arange(32, 64, 32))
class ModelOutput(BaseModel):
    success: bool
    error: Optional[str]
    segmentedImage: Optional[Path]
class Predictor(BasePredictor):

    def predict(self, height: int=Input(description='Height of image', default=128, choices=INPUT_DIMS)) -> ModelOutput:
        return None

class SchnellPredictor(Predictor):

    def predict(self, height: int=Input(description='Height of image', default=128, choices=INPUT_DIMS)) -> ModelOutput:
        return None"""
    ), "Stripped code needs to equal the minimum viable type inference."


@pytest.mark.skipif(sys.version_info < (3, 9), reason="Requires Python 3.9 or newer")
def test_strip_model_source_code_keeps_referenced_class_from_function():
    stripped_code = strip_model_source_code(
        """
from cog import BaseModel, Input, Path

class TrainingOutput(BaseModel):
    weights: Path

def train(
    n: int,
) -> TrainingOutput:
    with open("weights.bin", "w") as fh:
        for _ in range(n):
            fh.write("a")

    return TrainingOutput(
        weights=Path("weights.bin"),
    )
""",
        ["train"],
        [],
    )
    assert (
        stripped_code
        == """from cog import BaseModel, Input, Path
class TrainingOutput(BaseModel):
    weights: Path
def train(n: int) -> TrainingOutput:
    return None"""
    ), "Stripped code needs to equal the minimum viable type inference."



import sys
from mimetypes import MimeTypes

from cog.mimetypes_ext import install_mime_extensions


def test_webp_ext_support():
    # Assert on empty database.
    mt = MimeTypes(filenames=tuple())
    if sys.version_info < (3, 13):
        assert mt.guess_type("image.webp") == (None, None)

    install_mime_extensions(mt)
    assert mt.guess_type("image.webp") == ("image/webp", None)

    # Assert global override
    import mimetypes

    import cog  # noqa: F401

    assert mimetypes.guess_type("image.webp") == ("image/webp", None)



import importlib.util
import os
import os.path
import sys
from datetime import datetime
from enum import Enum
from types import ModuleType
from typing import Any, Dict, List, Optional, Type

import pydantic

from .types import PYDANTIC_V2

BUNDLED_SCHEMA_PATH = ".cog/schema.py"


class Status(str, Enum):
    STARTING = "starting"
    PROCESSING = "processing"
    SUCCEEDED = "succeeded"
    CANCELED = "canceled"
    FAILED = "failed"

    @staticmethod
    def is_terminal(status: Optional["Status"]) -> bool:
        return status in {Status.SUCCEEDED, Status.CANCELED, Status.FAILED}


class WebhookEvent(str, Enum):
    START = "start"
    OUTPUT = "output"
    LOGS = "logs"
    COMPLETED = "completed"

    @classmethod
    def default_events(cls) -> List["WebhookEvent"]:
        # if this is a set, it gets serialized to an array with an unstable ordering
        # so even though it's logically a set, have it as a list for deterministic schemas
        # note: this change removes "uniqueItems":true
        return [cls.START, cls.OUTPUT, cls.LOGS, cls.COMPLETED]


class PredictionBaseModel(pydantic.BaseModel):
    input: Dict[str, Any]

    if PYDANTIC_V2:
        model_config = pydantic.ConfigDict(use_enum_values=True)  # type: ignore
    else:

        class Config:
            # When using `choices`, the type is converted into an enum to validate
            # But, after validation, we want to pass the actual value to predict(), not the enum object
            use_enum_values = True


if PYDANTIC_V2:
    from pydantic.networks import UrlConstraints
    from pydantic_core import Url
    from typing_extensions import Annotated  # added to typing in python 3.9

    WebhookUrl = Annotated[
        Url, UrlConstraints(allowed_schemes=["http", "https"], max_length=65536)
    ]
else:
    WebhookUrl = pydantic.AnyUrl


class PredictionRequest(PredictionBaseModel):
    id: Optional[str] = None
    created_at: Optional[datetime] = None

    # TODO: deprecate this
    output_file_prefix: Optional[str] = None

    webhook: Optional[WebhookUrl] = None
    webhook_events_filter: Optional[List[WebhookEvent]] = pydantic.Field(
        default=WebhookEvent.default_events(),
    )

    @classmethod
    def with_types(cls, input_type: Type[Any]) -> Any:
        # [compat] Input is implicitly optional -- previous versions of the
        # Cog HTTP API allowed input to be omitted (e.g. for models that don't
        # have any inputs). We should consider changing this in future.
        return pydantic.create_model(
            cls.__name__, __base__=cls, input=(Optional[input_type], None)
        )


class PredictionResponse(PredictionBaseModel):
    output: Any = None

    id: Optional[str] = None
    version: Optional[str] = None

    created_at: Optional[datetime] = None
    started_at: Optional[datetime] = None
    completed_at: Optional[datetime] = None

    logs: str = ""
    error: Optional[str] = None
    status: Optional[Status] = None

    metrics: Optional[Dict[str, Any]] = None

    # This is used to track a fatal exception that occurs during a prediction.
    # "Fatal" means that we require the worker to be shut down to recover:
    # regular exceptions raised during predict are handled and do not use this
    # field.
    _fatal_exception: Optional[BaseException] = pydantic.PrivateAttr(default=None)

    @classmethod
    def with_types(cls, input_type: Type[Any], output_type: Type[Any]) -> Any:
        # [compat] Input is implicitly optional -- previous versions of the
        # Cog HTTP API allowed input to be omitted (e.g. for models that don't
        # have any inputs). We should consider changing this in future.
        return pydantic.create_model(
            cls.__name__,
            __base__=cls,
            input=(Optional[input_type], None),
            output=(Optional[output_type], None),
        )


class TrainingRequest(PredictionRequest):
    pass


class TrainingResponse(PredictionResponse):
    pass


def create_schema_module() -> Optional[ModuleType]:
    if not os.path.exists(BUNDLED_SCHEMA_PATH):
        return None
    name = "cog.bundled_schema"
    spec = importlib.util.spec_from_file_location(name, BUNDLED_SCHEMA_PATH)
    assert spec is not None
    module = importlib.util.module_from_spec(spec)
    assert spec.loader is not None
    sys.modules[name] = module
    spec.loader.exec_module(module)
    return module



import os
import tempfile

import numpy as np
import pydantic
import responses

import cog
from cog.files import upload_file
from cog.json import make_encodeable, upload_files
from cog.types import PYDANTIC_V2, URLFile


def test_make_encodeable_recursively_encodes_tuples():
    result = make_encodeable((np.float32(0.1), np.float32(0.2)))
    assert isinstance(result[0], float)


def test_make_encodeable_encodes_pydantic_models():
    class Model(pydantic.BaseModel):
        text: str
        number: int

        if PYDANTIC_V2:
            model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)
        else:

            class Config:
                arbitrary_types_allowed = True

    assert make_encodeable(Model(text="hello", number=5)) == {
        "text": "hello",
        "number": 5,
    }


def test_make_encodeable_ignores_files():
    class Model(pydantic.BaseModel):
        path: cog.Path

    temp_dir = tempfile.mkdtemp()
    temp_path = os.path.join(temp_dir, "my_file.txt")
    with open(temp_path, "w") as fh:
        fh.write("file content")
    path = cog.Path(temp_path)
    model = Model(path=path)
    assert make_encodeable(model) == {"path": path}


def test_upload_files():
    temp_dir = tempfile.mkdtemp()
    temp_path = os.path.join(temp_dir, "my_file.txt")
    with open(temp_path, "w") as fh:
        fh.write("file content")
    obj = {"path": cog.Path(temp_path)}
    assert upload_files(obj, upload_file) == {
        "path": "data:text/plain;base64,ZmlsZSBjb250ZW50"
    }


@responses.activate
def test_upload_files_with_url():
    responses.get(
        "https://example.com/some/url.txt",
        body="file content",
        status=200,
    )

    obj = {"path": URLFile("https://example.com/some/url.txt")}
    assert upload_files(obj, upload_file) == {
        "path": "data:text/plain;base64,ZmlsZSBjb250ZW50"
    }


def test_numpy():
    class Model(pydantic.BaseModel):
        ndarray: np.ndarray
        npfloat: np.float64
        npinteger: np.integer

        if PYDANTIC_V2:
            model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)
        else:

            class Config:
                arbitrary_types_allowed = True

    model = Model(
        ndarray=np.array([[1, 2], [3, 4]]),
        npfloat=np.float64(1.3),
        npinteger=np.int32(5),
    )
    assert make_encodeable(model) == {
        "ndarray": [[1, 2], [3, 4]],
        "npfloat": 1.3,
        "npinteger": 5,
    }



from cog import BasePredictor, File


class Predictor(BasePredictor):
    def predict(self, file: File) -> str:
        content = file.read()
        if isinstance(content, bytes):
            # Decode bytes to str assuming UTF-8 encoding; adjust if needed
            content = content.decode('utf-8')
        return content



import pathlib
import shutil
import subprocess
from pathlib import Path


def test_train_takes_input_and_produces_weights(tmpdir_factory):
    project_dir = Path(__file__).parent / "fixtures/train-project"
    out_dir = pathlib.Path(tmpdir_factory.mktemp("project"))
    shutil.copytree(project_dir, out_dir, dirs_exist_ok=True)
    result = subprocess.run(
        ["cog", "train", "--debug", "-i", "n=42"],
        cwd=out_dir,
        check=False,
        capture_output=True,
    )
    assert result.returncode == 0
    assert result.stdout == b""
    with open(out_dir / "weights.bin", "rb") as f:
        assert len(f.read()) == 42
    assert "falling back to slow loader" not in str(result.stderr)


def test_train_pydantic2(tmpdir_factory):
    project_dir = Path(__file__).parent / "fixtures/pydantic2-output"
    out_dir = pathlib.Path(tmpdir_factory.mktemp("project"))
    shutil.copytree(project_dir, out_dir, dirs_exist_ok=True)
    result = subprocess.run(
        ["cog", "train", "--debug", "-i", 'some_input="hello"'],
        cwd=out_dir,
        check=False,
        capture_output=True,
    )
    assert result.returncode == 0



import threading
from typing import Iterator

from cog import BasePredictor


def keep_printing():
    for _ in range(10000):
        print("hello")


class Predictor(BasePredictor):
    def predict(self) -> Iterator[str]:
        print_thread = threading.Thread(target=keep_printing)
        print_thread.start()
        yield "output" * 10000
        yield "output" * 10000
        yield "output" * 10000
        print_thread.join()



import tempfile

from cog import BasePredictor, Path


class Predictor(BasePredictor):
    def setup(self):
        self.foo = "foo"

    def predict(self, text: str, path: Path) -> Path:
        with open(path) as f:
            output = self.foo + text + f.read()
        tmpdir = Path(tempfile.mkdtemp())
        with open(tmpdir / "output.txt", "w") as fh:
            fh.write(output)
        return tmpdir / "output.txt"



import base64
import os
import sys
import threading
import time

import pytest
import responses
from werkzeug.wrappers import Response

from cog import schema
from cog.config import Config
from cog.server.http import Health, create_app
from cog.types import PYDANTIC_V2

from .conftest import _fixture_path, uses_predictor


@uses_predictor("input_none")
def test_no_input(client, match):
    resp = client.post("/predictions")
    assert resp.status_code == 200
    assert resp.json() == match({"status": "succeeded", "output": "foobar"})


@uses_predictor("input_none")
def test_missing_input(client, match):
    """Check we support missing input fields for backwards compatibility"""
    resp = client.post("/predictions", json={})
    assert resp.status_code == 200
    assert resp.json() == match({"status": "succeeded", "output": "foobar"})


@uses_predictor("input_none")
def test_empty_input(client, match):
    """Check we support empty input fields for backwards compatibility"""
    resp = client.post("/predictions", json={"input": {}})
    assert resp.status_code == 200
    assert resp.json() == match({"status": "succeeded", "output": "foobar"})


@uses_predictor("input_integer")
def test_good_int_input(client, match):
    resp = client.post("/predictions", json={"input": {"num": 3}})
    assert resp.status_code == 200
    assert resp.json() == match({"output": 27, "status": "succeeded"})
    resp = client.post("/predictions", json={"input": {"num": -3}})
    assert resp.status_code == 200
    assert resp.json() == match({"output": -27, "status": "succeeded"})


@uses_predictor("input_integer")
def test_bad_int_input(client):
    resp = client.post("/predictions", json={"input": {"num": "foo"}})
    detail = resp.json()["detail"][0]
    assert detail["loc"] == ["body", "input", "num"]
    assert "valid integer" in detail["msg"]

    assert resp.status_code == 422


@uses_predictor("input_integer_default")
def test_default_int_input(client, match):
    resp = client.post("/predictions", json={"input": {}})
    assert resp.status_code == 200
    assert resp.json() == match({"output": 25, "status": "succeeded"})

    resp = client.post("/predictions", json={"input": {"num": 3}})
    assert resp.status_code == 200
    assert resp.json() == match({"output": 9, "status": "succeeded"})


@uses_predictor("input_file")
def test_file_input_data_url(client, match):
    resp = client.post(
        "/predictions",
        json={
            "input": {
                "file": "data:text/plain;base64,"
                + base64.b64encode(b"bar").decode("utf-8")
            }
        },
    )
    assert resp.json() == match({"output": "bar", "status": "succeeded"})
    assert resp.status_code == 200


@uses_predictor("input_file")
def test_file_input_with_http_url(client, httpserver, match):
    # Use a real HTTP server rather than responses as file fetching occurs on
    # the other side of the Worker process boundary.
    httpserver.expect_request("/foo.txt").respond_with_data("hello")
    resp = client.post(
        "/predictions",
        json={"input": {"file": httpserver.url_for("/foo.txt")}},
    )
    assert resp.json() == match({"output": "hello", "status": "succeeded"})


@uses_predictor("input_path_2")
def test_file_input_with_http_url_error(client, httpserver, match):
    httpserver.expect_request("/foo.txt").respond_with_data("haha", status=404)
    resp = client.post(
        "/predictions",
        json={"input": {"path": httpserver.url_for("/foo.txt")}},
    )
    assert resp.json() == match({"status": "failed"})


@uses_predictor("input_path")
def test_path_input_data_url(client, match):
    resp = client.post(
        "/predictions",
        json={
            "input": {
                "path": "data:text/plain;base64,"
                + base64.b64encode(b"bar").decode("utf-8")
            }
        },
    )
    assert resp.json() == match({"output": "txt bar", "status": "succeeded"})
    assert resp.status_code == 200


@uses_predictor("input_path")
def test_path_input_slow_response(client, httpserver, match):
    def _handle(_):
        time.sleep(5)
        return Response("Slow response!")

    httpserver.expect_request("/foo.txt").respond_with_handler(_handle)
    now = time.monotonic()
    resp = client.post(
        "/predictions",
        json={
            "input": {
                "path": httpserver.url_for("/foo.txt"),
            }
        },
        headers={"Prefer": "respond-async"},
    )
    # The download of the slow input file should not happen during the request.
    assert time.monotonic() - now < 0.2
    assert resp.json() == match({"status": "processing"})
    assert resp.status_code == 202


@uses_predictor("input_path_2")
def test_path_temporary_files_are_removed(client, match):
    resp = client.post(
        "/predictions",
        json={
            "input": {
                "path": "data:text/plain;base64,"
                + base64.b64encode(b"bar").decode("utf-8")
            }
        },
    )
    temporary_path = resp.json()["output"]

    # HACK: the temp file is deleted in a concurrent.futures callback, which
    # isn't guaranteed to return before the future result resolves.  Therefore
    # the file might still exist at the point we receive the HTTP response.

    if not os.path.exists(temporary_path):
        return  # the file is gone, we're done
    # else wait and try again
    time.sleep(0.2)
    assert not os.path.exists(temporary_path)


@responses.activate
@uses_predictor("input_path")
def test_path_input_with_http_url(client, match):
    responses.add(responses.GET, "http://example.com/foo.txt", body="hello")
    resp = client.post(
        "/predictions",
        json={"input": {"path": "http://example.com/foo.txt"}},
    )
    assert resp.json() == match({"output": "txt hello", "status": "succeeded"})


@uses_predictor("input_file")
def test_file_bad_input(client):
    resp = client.post(
        "/predictions",
        json={"input": {"file": "foo"}},
    )
    assert resp.status_code == 422


@uses_predictor("input_multiple")
def test_multiple_arguments(client, match):
    resp = client.post(
        "/predictions",
        json={
            "input": {
                "text": "baz",
                "num1": 5,
                "path": "data:text/plain;base64,"
                + base64.b64encode(b"wibble").decode("utf-8"),
            }
        },
    )
    assert resp.status_code == 200
    assert resp.json() == match({"output": "baz 50 wibble", "status": "succeeded"})


@uses_predictor("input_ge_le")
def test_gt_lt(client):
    resp = client.post("/predictions", json={"input": {"num": 2}})
    detail = resp.json()["detail"][0]
    assert detail["loc"] == ["body", "input", "num"]
    assert "greater than or equal to 3.01" in detail["msg"]

    resp = client.post("/predictions", json={"input": {"num": 5}})
    assert resp.status_code == 200


@uses_predictor("input_choices")
def test_choices_str(client):
    resp = client.post("/predictions", json={"input": {"text": "foo"}})
    assert resp.status_code == 200
    resp = client.post("/predictions", json={"input": {"text": "baz"}})
    assert resp.status_code == 422


@uses_predictor("input_choices_iterable")
def test_choices_str_iterable(client):
    resp = client.post("/predictions", json={"input": {"text": "foo"}})
    assert resp.status_code == 200
    resp = client.post("/predictions", json={"input": {"text": "baz"}})
    assert resp.status_code == 422


@uses_predictor("input_choices_integer")
def test_choices_int(client):
    resp = client.post("/predictions", json={"input": {"x": 1}})
    assert resp.status_code == 200
    resp = client.post("/predictions", json={"input": {"x": 3}})
    assert resp.status_code == 422


@pytest.mark.skipif(
    not PYDANTIC_V2,
    reason="Literal is used for enums only in Pydantic v2",
)
@uses_predictor("input_literal")
def test_literal_str(client):
    resp = client.post("/predictions", json={"input": {"text": "foo"}})
    assert resp.status_code == 200
    resp = client.post("/predictions", json={"input": {"text": "baz"}})
    assert resp.status_code == 422


@pytest.mark.skipif(
    not PYDANTIC_V2,
    reason="Literal is used for enums only in Pydantic v2",
)
@uses_predictor("input_literal_integer")
def test_literal_int(client):
    resp = client.post("/predictions", json={"input": {"x": 1}})
    assert resp.status_code == 200
    resp = client.post("/predictions", json={"input": {"x": 3}})
    assert resp.status_code == 422


@uses_predictor("input_union_string_or_list_of_strings")
def test_union_strings(client):
    resp = client.post("/predictions", json={"input": {"args": "abc"}})
    assert resp.status_code == 200
    assert resp.json()["output"] == "abc"

    resp = client.post("/predictions", json={"input": {"args": ["a", "b", "c"]}})
    assert resp.status_code == 200
    assert resp.json()["output"] == "abc"

    # FIXME: Numbers are successfully cast to strings, but maybe shouldn't be
    # resp = client.post("/predictions", json={"input": {"args": 123}})
    # assert resp.status_code == 422
    # resp = client.post("/predictions", json={"input": {"args": [1, 2, 3]}})
    # assert resp.status_code == 422


@uses_predictor("input_union_integer_or_list_of_integers")
def test_union_integers(client):
    resp = client.post("/predictions", json={"input": {"args": 123}})
    assert resp.status_code == 200
    assert resp.json()["output"] == 123

    resp = client.post("/predictions", json={"input": {"args": [1, 2, 3]}})
    assert resp.status_code == 200
    assert resp.json()["output"] == 6

    resp = client.post("/predictions", json={"input": {"args": "abc"}})
    assert resp.status_code == 422
    resp = client.post("/predictions", json={"input": {"args": ["a", "b", "c"]}})
    assert resp.status_code == 422


@uses_predictor("input_secret")
def test_secret_str(client, match):
    resp = client.post("/predictions", json={"input": {"secret": "foo"}})
    assert resp.status_code == 200
    assert resp.json() == match({"output": "foo", "status": "succeeded"})

    resp = client.post("/predictions", json={"input": {"secret": {}}})
    assert resp.status_code == 422


def test_untyped_inputs():
    config = {"predict": _fixture_path("input_untyped")}
    app = create_app(
        cog_config=Config(config),
        shutdown_event=threading.Event(),
        upload_url="input_untyped",
    )
    assert app.state.health == Health.SETUP_FAILED
    assert app.state.setup_result.status == schema.Status.FAILED
    assert "TypeError: No input type provided for parameter" in "".join(
        app.state.setup_result.logs
    )


def test_input_with_unsupported_type():
    config = {"predict": _fixture_path("input_unsupported_type")}
    app = create_app(
        cog_config=Config(config),
        shutdown_event=threading.Event(),
        upload_url="input_untyped",
    )
    assert app.state.health == Health.SETUP_FAILED
    assert app.state.setup_result.status == schema.Status.FAILED
    assert "TypeError: Unsupported input type input_unsupported_type" in "".join(
        app.state.setup_result.logs
    )


@pytest.mark.skipif(sys.version_info < (3, 10), reason="Requires Python 3.10 or newer")
@uses_predictor("input_path_or_none")
def test_path_or_none(client, httpserver, match):
    httpserver.expect_request("/foo.txt").respond_with_data("hello")
    resp = client.post(
        "/predictions",
        json={"input": {"file": httpserver.url_for("/foo.txt")}},
    )
    assert resp.json() == match({"output": "hello", "status": "succeeded"})



from cog import BasePredictor
from pydantic import BaseModel


# Calling this `MyOutput` to test if cog renames it to `Output` in the schema
class MyOutput(BaseModel):
    foo_number: int = "42"
    foo_string: str = "meaning of life"


class Predictor(BasePredictor):
    def predict(
        self,
    ) -> MyOutput:
        pass



import asyncio

from cog import BasePredictor


class Predictor(BasePredictor):
    async def predict(self, sleep: float = 0) -> str:
        print("starting")
        await asyncio.sleep(sleep)
        return f"done in {sleep} seconds"



import os
from typing import Any, Callable, Set

import requests
import structlog
from fastapi.encoders import jsonable_encoder
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from ..schema import PredictionResponse, Status, WebhookEvent
from ..types import PYDANTIC_V2
from .response_throttler import ResponseThrottler
from .telemetry import current_trace_context
from .useragent import get_user_agent

log = structlog.get_logger(__name__)

_response_interval = float(os.environ.get("COG_THROTTLE_RESPONSE_INTERVAL", 0.5))

# HACK: signal that we should skip the start webhook when the response interval
# is tuned below 100ms. This should help us get output sooner for models that
# are latency sensitive.
SKIP_START_EVENT = _response_interval < 0.1


def webhook_caller_filtered(
    webhook: str,
    webhook_events_filter: Set[WebhookEvent],
) -> Callable[[Any, WebhookEvent], None]:
    upstream_caller = webhook_caller(webhook)

    def caller(response: PredictionResponse, event: WebhookEvent) -> None:
        if event in webhook_events_filter:
            upstream_caller(response)

    return caller


def webhook_caller(webhook: str) -> Callable[[Any], None]:
    # TODO: we probably don't need to create new sessions and new throttlers
    # for every prediction.
    throttler = ResponseThrottler(response_interval=_response_interval)

    default_session = requests_session()
    retry_session = requests_session_with_retries()

    def caller(response: PredictionResponse) -> None:
        if throttler.should_send_response(response):
            if PYDANTIC_V2:
                dict_response = jsonable_encoder(
                    response.model_dump(exclude_unset=True)
                )
            else:
                dict_response = jsonable_encoder(response.dict(exclude_unset=True))
            if Status.is_terminal(response.status):
                # For terminal updates, retry persistently
                retry_session.post(webhook, json=dict_response)
            else:
                # For other requests, don't retry, and ignore any errors
                try:
                    default_session.post(webhook, json=dict_response)
                except requests.exceptions.RequestException:
                    log.warn("caught exception while sending webhook", exc_info=True)
            throttler.update_last_sent_response_time()

    return caller


def requests_session() -> requests.Session:
    session = requests.Session()
    session.headers["user-agent"] = (
        get_user_agent() + " " + str(session.headers["user-agent"])
    )

    ctx = current_trace_context() or {}
    for key, value in ctx.items():
        session.headers[key] = str(value)

    auth_token = os.environ.get("WEBHOOK_AUTH_TOKEN")
    if auth_token:
        session.headers["authorization"] = "Bearer " + auth_token

    return session


def requests_session_with_retries() -> requests.Session:
    # This session will retry requests up to 12 times, with exponential
    # backoff. In total it'll try for up to roughly 320 seconds, providing
    # resilience through temporary networking and availability issues.
    session = requests_session()
    adapter = HTTPAdapter(
        max_retries=Retry(
            total=12,
            backoff_factor=0.1,
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=["POST"],
        )
    )
    session.mount("http://", adapter)
    session.mount("https://", adapter)

    return session



from anthropic.types import MessageParam
from cog import BasePredictor, Input
from cog.coder import json_coder  # noqa: F401


class Predictor(BasePredictor):
    def predict(
        self,
        message: MessageParam = Input(description="Messages API."),
    ) -> str:
        return "Content: " + message["content"]



class CogServerError(Exception):
    """Base class for all Cog server errors."""

    def __str__(self) -> str:
        return f"Cog: {super().__str__()}"


class FileUploadError(CogServerError):
    pass


class RunnerBusyError(CogServerError):
    pass


class UnknownPredictionError(CogServerError):
    pass


class CogRuntimeError(CogServerError, RuntimeError):
    pass


class CogTimeoutError(CogServerError, TimeoutError):
    pass



import os
import tempfile

from cog import BasePredictor, Path
from PIL import Image


class Predictor(BasePredictor):
    def predict(self) -> Path:
        temp_dir = tempfile.mkdtemp()
        temp_path = os.path.join(temp_dir, "my_file.bmp")
        img = Image.new("RGB", (255, 255), "red")
        img.save(temp_path)
        return Path(temp_path)



import sys

sys.stdout.write("writing to stdout at import time\n")
sys.stderr.write("writing to stderr at import time\n")

import missing_module


class Predictor:
    def setup(self):
        pass

    def predict(self):
        print("did predict")



from typing import Iterator, List

from cog import BasePredictor
from pydantic import BaseModel


class Output(BaseModel):
    text: str


class Predictor(BasePredictor):
    def predict(self) -> Iterator[List[Output]]:
        yield [Output(text="hello")]



import inspect
import os
import sys
from typing import Optional
from unittest.mock import patch

from pydantic.fields import FieldInfo

from cog import File, Input, Path
from cog.predictor import (
    get_input_create_model_kwargs,
    get_predict,
    get_weights_type,
    load_predictor_from_ref,
)
from cog.types import PYDANTIC_V2

if PYDANTIC_V2:
    from pydantic.fields import PydanticUndefined
else:
    from pydantic.fields import Undefined as PydanticUndefined


def is_field_required(field: FieldInfo):
    if hasattr(field, "is_required"):
        return field.is_required()
    if hasattr(field, "required"):
        return field.required
    return field.default is PydanticUndefined and field.default_factory is None


def test_get_weights_type() -> None:
    def f() -> None:
        pass

    assert get_weights_type(f) is None

    def f(weights: File) -> None:
        pass

    assert get_weights_type(f) == File

    def f(weights: Path) -> None:
        pass

    assert get_weights_type(f) == Path

    def f(weights: Optional[File]) -> None:
        pass

    assert get_weights_type(f) == File


def test_load_predictor_from_ref_overrides_argv():
    with patch("sys.argv", ["foo.py", "exec", "--giraffes=2", "--eat-cookies"]):
        predictor = load_predictor_from_ref(_fixture_path("argv_override"))

        # check the predictor module saw no args
        assert predictor.predict() == ["foo.py"]
        # check we reset the args correctly
        assert sys.argv == ["foo.py", "exec", "--giraffes=2", "--eat-cookies"]


def test_get_input_create_model_kwargs():
    def predict(thing: Optional[str] = Input(description="Hello String.")) -> str:
        return thing if thing is not None else "Nothing"

    predict_type = get_predict(predict)
    signature = inspect.signature(predict_type)
    output = get_input_create_model_kwargs(signature)
    assert not is_field_required(output["thing"][1])


def _fixture_path(name):
    test_dir = os.path.dirname(os.path.realpath(__file__))
    return os.path.join(test_dir, f"fixtures/{name}.py") + ":Predictor"



import signal
import subprocess
import sys

import requests

from cog import BasePredictor


class Predictor(BasePredictor):
    """
    This predictor checks the case where a process is spawned during setup and then each
    prediction depends on being able to communicate with that process. In the event that
    stream redirection is not working correctly, the forked process will not be able to
    write to stdout/stderr and will likely exit. Any state other than "running" is
    considered an error condition and raises SystemExit to interrupt any more prediction
    serving.

    This variant runs a forked python HTTP server via a shell wrapper to which a request
    is made during each call to `predict`.
    """

    def setup(self) -> None:
        print("---> starting background process")

        self.bg = subprocess.Popen(["bash", "run-pong.sh"])

        print(f"---> started background process pid={self.bg.pid}")

    def predict(self, s: str) -> str:
        status = self.bg.poll()

        print(f"---> background job status={status}")

        if status is None:
            print(f"---> sending request to background job pid={self.bg.pid}")

            print(requests.get("http://127.0.0.1:7777/ping"))

            print(f"---> sent request to background job pid={self.bg.pid}")
        else:
            raise SystemExit

        return "hello " + s



import os
import tempfile
from typing import Iterator

from cog import BasePredictor, Path
from PIL import Image


class Predictor(BasePredictor):
    def predict(self) -> Iterator[Path]:
        colors = ["red", "blue", "yellow"]
        for i, color in enumerate(colors):
            temp_dir = tempfile.mkdtemp()
            temp_path = os.path.join(temp_dir, f"prediction-{i}.bmp")
            img = Image.new("RGB", (255, 255), color)
            img.save(temp_path)
            yield Path(temp_path)



import os
import sys
import threading
import time
from contextlib import ExitStack
from typing import Any, Dict, Optional, Sequence, Tuple
from unittest import mock

import pytest
from attrs import define
from fastapi.testclient import TestClient

from cog.command import ast_openapi_schema
from cog.config import Config
from cog.server.http import create_app
from cog.server.worker import make_worker


@define
class AppConfig:
    predictor_fixture: str
    options: Optional[Dict[str, Any]]


@define
class WorkerConfig:
    fixture_name: str
    is_async: bool = False
    setup: bool = True
    max_concurrency: int = 1
    min_python: Optional[Tuple[int, int]] = None


def pytest_make_parametrize_id(config, val):
    """
    Generates more readable IDs for parametrized tests that use AppConfig or
    WorkerConfig values.
    """
    if isinstance(val, AppConfig):
        return val.predictor_fixture
    elif isinstance(val, WorkerConfig):
        return val.fixture_name


def _fixture_path(name):
    # HACK: `name` can either be in the form "<name>.py:Predictor" or just "<name>".
    if ":" not in name:
        name = f"{name}.py:Predictor"

    test_dir = os.path.dirname(os.path.realpath(__file__))
    return os.path.join(test_dir, f"fixtures/{name}")


def uses_predictor(name):
    return pytest.mark.parametrize(
        "client", [AppConfig(predictor_fixture=name, options={})], indirect=True
    )


def uses_trainer(name):
    # HACK: `name` can either be in the form "<name>.py:train" or just "<name>".
    if ":" not in name:
        name = f"{name}.py:train"
    options = {"additional_config": {"train": _fixture_path(name)}}

    return pytest.mark.parametrize(
        "client", [AppConfig(predictor_fixture=name, options=options)], indirect=True
    )


def uses_predictor_with_client_options(name, **options):
    return pytest.mark.parametrize(
        "client", [AppConfig(predictor_fixture=name, options=options)], indirect=True
    )


def uses_worker(
    name_or_names, setup=True, max_concurrency=1, min_python=None, is_async=False
):
    """
    Decorator for tests that require a Worker instance. `name_or_names` can be
    a single fixture name, or a sequence (list, tuple) of fixture names. If
    it's a sequence, the test will be run once for each worker.

    If `setup` is True (the default) setup will be run before the test runs.
    """
    if isinstance(name_or_names, (tuple, list)):
        values = (
            WorkerConfig(
                fixture_name=n,
                setup=setup,
                max_concurrency=max_concurrency,
                min_python=min_python,
                is_async=is_async,
            )
            for n in name_or_names
        )
    else:
        values = (
            WorkerConfig(
                fixture_name=name_or_names,
                setup=setup,
                max_concurrency=max_concurrency,
                min_python=min_python,
                is_async=is_async,
            ),
        )
    return uses_worker_configs(list(values))


def uses_worker_configs(values: Sequence[WorkerConfig]):
    """
    Decorator for tests that require a Worker instance.  The test will be
    run once for each worker.  `configs` is a sequence (list, tuple, generator)
    of WorkerConfig.
    """
    return pytest.mark.parametrize("worker", values, indirect=True)


def make_client(
    fixture_name: str,
    upload_url: Optional[str] = None,
    additional_config: Optional[dict] = None,
):
    """
    Creates a fastapi test client for an app that uses the requested Predictor.
    """

    config = {"predict": _fixture_path(fixture_name)}
    if additional_config:
        config.update(additional_config)

    app = create_app(
        cog_config=Config(config=config),
        shutdown_event=threading.Event(),
        upload_url=upload_url,
    )
    return TestClient(app)


def wait_for_setup(client: TestClient):
    while True:
        resp = client.get("/health-check")
        data = resp.json()
        if data["status"] != "STARTING":
            break
        time.sleep(0.01)


@pytest.fixture
def client(request):
    fixture_name = request.param.predictor_fixture
    options = request.param.options

    with ExitStack() as stack:
        if "env" in options:
            stack.enter_context(mock.patch.dict(os.environ, options["env"]))
            del options["env"]

        # Use context manager to trigger setup/shutdown events.
        c = make_client(fixture_name=fixture_name, **options)
        stack.enter_context(c)
        wait_for_setup(c)
        c.ref = fixture_name
        yield c


@pytest.fixture
def static_schema(client) -> dict:
    ref = _fixture_path(client.ref)
    module_path = ref.split(":", 1)[0]
    return ast_openapi_schema.extract_file(module_path)


@pytest.fixture
def worker(request):
    ref = _fixture_path(request.param.fixture_name)
    if (
        request.param.min_python is not None
        and sys.version_info < request.param.min_python
    ):
        pytest.skip(
            f"Test requires python {request.param.min_python[0]}.{request.param.min_python[1]}"
        )
    w = make_worker(
        predictor_ref=ref,
        is_async=request.param.is_async,
        tee_output=False,
        max_concurrency=request.param.max_concurrency,
    )
    if request.param.setup:
        assert not w.setup().result().error
    try:
        yield w
    finally:
        w.shutdown()



import os
import sys
from contextlib import contextmanager
from typing import Iterator


@contextmanager
def suppress_output() -> Iterator[None]:
    null_out = open(os.devnull, "w", encoding="utf-8")
    null_err = open(os.devnull, "w", encoding="utf-8")
    out_fd = sys.stdout.fileno()
    err_fd = sys.stderr.fileno()
    out_dup_fd = os.dup(out_fd)
    err_dup_fd = os.dup(err_fd)
    os.dup2(null_out.fileno(), out_fd)
    os.dup2(null_err.fileno(), err_fd)

    try:
        yield
    finally:
        os.dup2(out_dup_fd, out_fd)
        os.dup2(err_dup_fd, err_fd)
        null_out.close()
        null_err.close()
        os.close(out_dup_fd)
        os.close(err_dup_fd)



import time

from ..schema import PredictionResponse, Status


class ResponseThrottler:
    def __init__(self, response_interval: float) -> None:
        self.last_sent_response_time = 0.0
        self.response_interval = response_interval

    def should_send_response(self, response: PredictionResponse) -> bool:
        if Status.is_terminal(response.status):
            return True

        return self.seconds_since_last_response() >= self.response_interval

    def update_last_sent_response_time(self) -> None:
        self.last_sent_response_time = time.time()

    def seconds_since_last_response(self) -> float:
        return time.time() - self.last_sent_response_time



import os
import signal
import time
from random import randint
from wsgiref.simple_server import make_server


def main():
    child_pid = os.fork()
    is_child = child_pid == 0

    pid = os.getpid()

    if is_child:
        make_server("127.0.0.1", 7777, app).serve_forever()
    else:
        while True:
            print(f"===> PARENT ({pid})")

            time.sleep(10)


def app(environ, start_response):
    print(f"---> CHILD ({os.getpid()})")

    if environ["PATH_INFO"] == "/ping":
        start_response("200 OK", [("content-type", "text/plain")])
        return [b"PONG\n" for n in range(100 + randint(2, 32))]

    start_response("404 Not Found", [("content-type", "text/plain")])
    return [b"NO\n"]


if __name__ == "__main__":
    main()



import mimetypes

from pydantic import BaseModel

from .base_predictor import BasePredictor
from .mimetypes_ext import install_mime_extensions
from .server.scope import current_scope
from .types import (
    AsyncConcatenateIterator,
    ConcatenateIterator,
    ExperimentalFeatureWarning,
    File,
    Input,
    Path,
    Secret,
)

install_mime_extensions(mimetypes)

try:
    from ._version import __version__
except ImportError:
    __version__ = "0.0.0+unknown"


__all__ = [
    "__version__",
    "current_scope",
    "AsyncConcatenateIterator",
    "BaseModel",
    "BasePredictor",
    "ConcatenateIterator",
    "ExperimentalFeatureWarning",
    "File",
    "Input",
    "Path",
    "Secret",
]



from cog import BasePredictor, Path, Input
from PIL import Image


class Predictor(BasePredictor):
    def predict(self,test_image: Path | None = Input(description="Test image", default=None)) -> Path:
        """Run a single prediction on the model"""
        im = Image.new("RGB", (100, 100), color="red")
        im.save(Path("./hello.webp"))
        return Path("./hello.webp")
    


from cog import BasePredictor
from pydantic import BaseModel


class Input(BaseModel):
    text: str


class Predictor(BasePredictor):
    def predict(self, input: Input) -> str:
        return input.text



"""
python -m cog.command.specification

This prints a JSON object describing the inputs of the model.
"""

import json
from typing import Any, Dict, List, Union

from ..config import Config
from ..errors import CogError, ConfigDoesNotExist, PredictorNotSet
from ..schema import Status
from ..server.http import create_app
from ..suppress_output import suppress_output


def remove_title_next_to_ref(
    schema_node: Union[Dict[str, Any], List[Any]],
) -> Union[Dict[str, Any], List[Any]]:
    """
    Recursively remove 'title' from schema components that have a '$ref'.
    This function addresses a non-compliance issue in FastAPI's OpenAPI schema generation, where
    'title' fields adjacent to '$ref' fields can cause validation problems with some OpenAPI tools.
    """
    if isinstance(schema_node, dict):
        if "$ref" in schema_node and "title" in schema_node:
            del schema_node["title"]
        for _key, value in schema_node.items():
            remove_title_next_to_ref(value)
    elif isinstance(schema_node, list):  # type: ignore[reportUnnecessaryIsInstance]
        for i, item in enumerate(schema_node):
            schema_node[i] = remove_title_next_to_ref(item)
    return schema_node


if __name__ == "__main__":
    schema = {}
    try:
        with suppress_output():
            app = create_app(cog_config=Config(), shutdown_event=None, is_build=True)
            if (
                app.state.setup_result
                and app.state.setup_result.status == Status.FAILED
            ):
                raise CogError(app.state.setup_result.logs)
            schema = remove_title_next_to_ref(app.openapi())
    except ConfigDoesNotExist:
        raise ConfigDoesNotExist("no cog.yaml found or present") from None
    except PredictorNotSet:
        raise PredictorNotSet("no predict method found in Predictor") from None
    print(json.dumps(schema, indent=2))



from cog import BasePredictor, Input


class Predictor(BasePredictor):
    def predict(
        self, num: int = Input(description="Number of things", default=1, ge=2, le=10)
    ) -> int:
        return num * 2



import contextlib
import random
import re
import signal
import socket
import string
import subprocess
import sys
import time

import httpx
from packaging.version import VERSION_PATTERN

# From the SemVer spec: https://semver.org/
SEMVER_PATTERN = r"^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"


# Used to help ensure that the cog binary reports a semver version that matches
# the PEP440 version of the embedded Python package.
#
# These are all valid pairs:
#
#   SEMVER            PEP440           NOTES
#   0.11.2            0.11.2
#   0.11.2-alpha2     0.11.2a2         prerelease counters are not checked
#   0.11.2-beta1      0.11.2b1             "          "      "   "     "
#   0.11.2-rc4        0.11.2rc4            "          "      "   "     "
#   0.11.2-dev        0.11.2rc4.dev10  dev status overrides prerelease status
#   0.11.2+gabcd      0.11.2+gabce
#
# The following are not valid pairs:
#
#   SEMVER            PEP440           NOTES
#   0.11.2            0.11.3           mismatched release versions
#   0.11.2-alpha2     0.11.2alpha2     PEP440 uses 'a' instead of 'alpha'
#   0.11.2-alpha2     0.11.2b2         mismatched prerelease status
#   0.11.2-rc4        0.11.2rc4.dev10  dev status should have overridden prerelease status
#   0.11.2+gabcd      0.11.2+gdefg     mismatched local/build metadata
#
def assert_versions_match(semver_version: str, pep440_version: str):
    semver_re = re.compile(SEMVER_PATTERN)
    pep440_re = re.compile(VERSION_PATTERN, re.VERBOSE | re.IGNORECASE)

    semver_match = semver_re.match(semver_version)
    pep440_match = pep440_re.match(pep440_version)

    assert semver_match, f"Invalid semver version: {semver_version}"
    assert pep440_match, f"Invalid PEP 440 version: {pep440_version}"

    semver_groups = semver_match.groupdict()
    pep440_groups = pep440_match.groupdict()

    semver_release = (
        f"{semver_groups['major']}.{semver_groups['minor']}.{semver_groups['patch']}"
    )

    # Check base release version
    assert semver_release == pep440_groups["release"], (
        f"Release versions do not match: {semver_release} != {pep440_groups['release']}"
    )

    # Check prerelease status
    semver_pre = semver_groups["prerelease"]
    pep440_pre = pep440_groups["pre"] or pep440_groups["dev"]

    assert bool(semver_pre) == bool(pep440_pre), "Pre-release status does not match"

    if semver_pre:
        if semver_pre.startswith("alpha"):
            assert pep440_groups["pre_l"] == "a", (
                "Alpha pre-release status does not match"
            )
            assert not pep440_groups["dev"], (
                "Semver pre-release cannot also be a PEP440 dev build"
            )

        if semver_pre.startswith("beta"):
            assert pep440_groups["pre_l"] == "b", (
                "Beta pre-release status does not match"
            )
            assert not pep440_groups["dev"], (
                "Semver pre-release cannot also be a PEP440 dev build"
            )

        if semver_pre.startswith("rc"):
            assert pep440_groups["pre_l"] == "rc", (
                "Release candidate pre-release status does not match"
            )
            assert not pep440_groups["dev"], (
                "Semver pre-release cannot also be a PEP440 dev build"
            )

        if semver_pre.startswith("dev"):
            assert pep440_groups["dev_l"] == "dev", "Dev build status does not match"

    assert semver_groups["buildmetadata"] == pep440_groups["local"], (
        f"Local/build metadata component does not match: {semver_groups['buildmetadata']} != {pep440_groups['local']}"
    )


def random_string(length):
    return "".join(random.choice(string.ascii_lowercase) for i in range(length))


def remove_docker_image(image_name, max_attempts=5, wait_seconds=1):
    for attempt in range(max_attempts):
        try:
            subprocess.run(
                ["docker", "rmi", "-f", image_name], check=True, capture_output=True
            )
            print(f"Image {image_name} successfully removed.")
            break
        except subprocess.CalledProcessError as e:
            print(f"Attempt {attempt + 1} failed: {e.stderr.decode()}")
            time.sleep(wait_seconds)
    else:
        print(f"Failed to remove image {image_name} after {max_attempts} attempts.")


def random_port() -> int:
    sock = socket.socket()
    sock.bind(("127.0.0.1", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


@contextlib.contextmanager
def cog_server_http_run(project_dir: str):
    port = random_port()
    addr = f"http://127.0.0.1:{port}"

    server: subprocess.Popen | None = None

    try:
        server = subprocess.Popen(
            [
                "cog",
                "serve",
                "-p",
                str(port),
            ],
            cwd=project_dir,
            # NOTE: inheriting stdout and stderr from the parent process when running
            # within a pytest context seems to *always fail*, even when using
            # `capsys.disabled` or `--capture=no` via command line args. Piping the
            # streams seems to allow behavior that is identical to code run outside of
            # pytest.
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        i = 0

        while True:
            try:
                if httpx.get(f"{addr}/health-check").status_code == 200:
                    break
            except httpx.HTTPError:
                pass

            time.sleep((0.1 + i) * 2)
            i += 1

        yield addr
    finally:
        try:
            httpx.post(f"{addr}/shutdown")
        except httpx.HTTPError:
            pass

        if server is not None:
            server.send_signal(signal.SIGINT)

            out, err = server.communicate(timeout=5)

            if server.returncode != 0:
                print(out.decode())
                print(err.decode(), file=sys.stderr)



from cog import BasePredictor, File, Input, Path


class Predictor(BasePredictor):
    def predict(
        self,
        no_default: str,
        default_without_input: str = "default",
        input_with_default: int = Input(default=-10),
        path: Path = Input(description="Some path"),
        image: File = Input(description="Some path"),
        choices: str = Input(choices=["foo", "bar"]),
        int_choices: int = Input(choices=[3, 4, 5]),
    ) -> str:
        pass



import subprocess


def test_run(tmpdir_factory):
    tmpdir = tmpdir_factory.mktemp("project")
    with open(tmpdir / "cog.yaml", "w") as f:
        cog_yaml = """
build:
  python_version: "3.8"
        """
        f.write(cog_yaml)

    result = subprocess.run(
        ["cog", "run", "echo", "hello world"],
        cwd=tmpdir,
        check=True,
        capture_output=True,
    )
    assert b"hello world" in result.stdout


def test_run_with_secret(tmpdir_factory):
    tmpdir = tmpdir_factory.mktemp("project")
    with open(tmpdir / "cog.yaml", "w") as f:
        cog_yaml = """
build:
  python_version: "3.8"
  run:
    - echo hello world
    - command: >-
        echo shh
      mounts:
        - type: secret
          id: foo
          target: secret.txt
        """
        f.write(cog_yaml)
    with open(tmpdir / "secret.txt", "w") as f:
        f.write("🤫")

    result = subprocess.run(
        ["cog", "debug"],
        cwd=tmpdir,
        check=True,
        capture_output=True,
    )
    assert b"RUN echo hello world" in result.stdout
    assert b"RUN --mount=type=secret,id=foo,target=secret.txt echo shh" in result.stdout



import time

from cog.schema import PredictionResponse, Status
from cog.server.response_throttler import ResponseThrottler

processing = PredictionResponse(input={}, status=Status.PROCESSING)
succeeded = PredictionResponse(input={}, status=Status.SUCCEEDED)


def test_zero_interval():
    throttler = ResponseThrottler(response_interval=0)

    assert throttler.should_send_response(processing)
    throttler.update_last_sent_response_time()
    assert throttler.should_send_response(succeeded)


def test_terminal_status():
    throttler = ResponseThrottler(response_interval=10)

    assert throttler.should_send_response(processing)
    throttler.update_last_sent_response_time()
    assert not throttler.should_send_response(processing)
    throttler.update_last_sent_response_time()
    assert throttler.should_send_response(succeeded)


def test_nonzero_internal():
    throttler = ResponseThrottler(response_interval=0.2)

    assert throttler.should_send_response(processing)
    throttler.update_last_sent_response_time()
    assert not throttler.should_send_response(processing)
    throttler.update_last_sent_response_time()

    time.sleep(0.3)

    assert throttler.should_send_response(processing)



from abc import ABC, abstractmethod
from typing import Any, Optional

from .types import Weights


class BasePredictor(ABC):
    def setup(
        self,
        weights: Optional[Weights] = None,  # pylint: disable=unused-argument
    ) -> None:
        """
        An optional method to prepare the model so multiple predictions run efficiently.
        """
        return

    @abstractmethod
    def predict(self, **kwargs: Any) -> Any:
        """
        Run a single prediction on the model
        """



import io
import mimetypes
import os
import pathlib
import shutil
import tempfile
import urllib.parse
import urllib.request
import urllib.response
from typing import (
    Any,
    AsyncIterator,
    Dict,
    Iterator,
    List,
    Optional,
    Type,
    TypedDict,
    TypeVar,
    Union,
)

import pydantic
import requests
from typing_extensions import NotRequired  # added to typing in python 3.11

if pydantic.__version__.startswith("1."):
    PYDANTIC_V2 = False
else:
    PYDANTIC_V2 = True


FILENAME_ILLEGAL_CHARS = set("\u0000/")

# Linux allows files up to 255 bytes long. We enforce a slightly shorter
# filename so that there's room for prefixes added by
# tempfile.NamedTemporaryFile, etc.
FILENAME_MAX_LENGTH = 200


class ExperimentalFeatureWarning(Warning):
    pass


class CogConfig(TypedDict):  # pylint: disable=too-many-ancestors
    build: "CogBuildConfig"
    concurrency: "CogConcurrencyConfig"
    image: NotRequired[str]
    predict: NotRequired[str]
    train: NotRequired[str]


class CogBuildConfig(TypedDict, total=False):  # pylint: disable=too-many-ancestors
    cuda: Optional[str]
    gpu: Optional[bool]
    python_packages: Optional[List[str]]
    system_packages: Optional[List[str]]
    python_requirements: Optional[str]
    python_version: Optional[str]
    run: Optional[Union[List[str], List[Dict[str, Any]]]]


class CogConcurrencyConfig(TypedDict, total=False):  # pylint: disable=too-many-ancestors
    max: NotRequired[int]


def Input(  # pylint: disable=invalid-name, too-many-arguments
    default: Any = ...,
    description: Optional[str] = None,
    ge: Optional[float] = None,
    le: Optional[float] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
    choices: Optional[List[Union[str, int]]] = None,
) -> Any:
    """Input is similar to pydantic.Field, but doesn't require a default value to be the first argument."""
    field_kwargs = {
        "default": default,
        "description": description,
        "ge": ge,
        "le": le,
        "min_length": min_length,
        "max_length": max_length,
    }

    if PYDANTIC_V2:
        field_kwargs["pattern"] = regex
        if choices:
            # The `choices` parameter is deprecated in Pydantic v2.
            # Instead, the user should use `Literal[...]`
            # to specify the allowed values.
            field_kwargs["json_schema_extra"] = {"enum": choices}
    else:
        field_kwargs["regex"] = regex
        field_kwargs["enum"] = choices
    return pydantic.Field(**field_kwargs)


class Secret(pydantic.SecretStr):
    if PYDANTIC_V2:
        from pydantic.json_schema import JsonSchemaValue
        from pydantic_core import CoreSchema

        @classmethod
        def __get_pydantic_json_schema__(
            cls, core_schema: CoreSchema, handler: Any
        ) -> JsonSchemaValue:
            json_schema = handler(core_schema)
            json_schema.update(
                {
                    "type": "string",
                    "format": "password",
                    "x-cog-secret": True,
                }
            )
            return json_schema

    else:

        @classmethod
        def __modify_schema__(cls, field_schema: Dict[str, Any]) -> None:
            """Defines what this type should be in openapi.json"""
            field_schema.update(
                {
                    "type": "string",
                    "format": "password",
                    "x-cog-secret": True,
                }
            )


class File(io.IOBase):
    """Deprecated: use Path instead."""

    validate_always = True

    @classmethod
    def validate(cls, value: Any) -> io.IOBase:
        if isinstance(value, io.IOBase):
            return value

        parsed_url = urllib.parse.urlparse(value)
        if parsed_url.scheme == "data":
            with urllib.request.urlopen(value) as res:  # noqa: S310
                return io.BytesIO(res.read())
        if parsed_url.scheme in ("http", "https"):
            return URLFile(value)
        raise ValueError(
            f"'{parsed_url.scheme}' is not a valid URL scheme. 'data', 'http', or 'https' is supported."
        )

    if PYDANTIC_V2:
        from pydantic import GetCoreSchemaHandler, TypeAdapter
        from pydantic_core.core_schema import CoreSchema

        @classmethod
        def __get_pydantic_core_schema__(
            cls,
            source: Type[Any],  # pylint: disable=unused-argument
            handler: "pydantic.GetCoreSchemaHandler",  # pylint: disable=unused-argument
        ) -> "CoreSchema":
            from pydantic_core import (  # pylint: disable=import-outside-toplevel
                core_schema,
            )

            return core_schema.union_schema(
                [
                    core_schema.is_instance_schema(io.IOBase),
                    core_schema.no_info_plain_validator_function(cls.validate),
                ]
            )

        @classmethod
        def __get_pydantic_json_schema__(
            cls, core_schema: "CoreSchema", handler: "pydantic.GetJsonSchemaHandler"
        ) -> "JsonSchemaValue":  # type: ignore # noqa: F821
            json_schema = handler(core_schema)
            json_schema.update(type="string", format="uri")
            return json_schema

    else:

        @classmethod
        def __get_validators__(cls) -> Iterator[Any]:
            yield cls.validate

        @classmethod
        def __modify_schema__(cls, field_schema: Dict[str, Any]) -> None:
            """Defines what this type should be in openapi.json"""
            # https://json-schema.org/understanding-json-schema/reference/string.html#uri-template
            field_schema.update(type="string", format="uri")


class Path(pathlib.PosixPath):  # pylint: disable=abstract-method
    validate_always = True

    @classmethod
    def validate(cls, value: Any) -> pathlib.Path:
        if isinstance(value, pathlib.Path):
            return value

        return URLPath(
            source=value,
            filename=get_filename(value),
            fileobj=File.validate(value),
        )

    if PYDANTIC_V2:
        from pydantic import GetCoreSchemaHandler
        from pydantic.json_schema import JsonSchemaValue
        from pydantic_core import CoreSchema

        @classmethod
        def __get_pydantic_core_schema__(
            cls,
            source: Type[Any],  # pylint: disable=unused-argument
            handler: "pydantic.GetCoreSchemaHandler",  # pylint: disable=unused-argument
        ) -> "CoreSchema":
            from pydantic_core import (  # pylint: disable=import-outside-toplevel
                core_schema,
            )

            return core_schema.union_schema(
                [
                    core_schema.is_instance_schema(pathlib.Path),
                    core_schema.no_info_plain_validator_function(cls.validate),
                ]
            )

        @classmethod
        def __get_pydantic_json_schema__(
            cls, core_schema: "CoreSchema", handler: "pydantic.GetJsonSchemaHandler"
        ) -> "JsonSchemaValue":  # type: ignore # noqa: F821
            json_schema = handler(core_schema)
            json_schema.update(type="string", format="uri")
            return json_schema

    else:

        @classmethod
        def __get_validators__(cls) -> Iterator[Any]:
            yield cls.validate

        @classmethod
        def __modify_schema__(cls, field_schema: Dict[str, Any]) -> None:
            """Defines what this type should be in openapi.json"""
            # https://json-schema.org/understanding-json-schema/reference/string.html#uri-template
            field_schema.update(type="string", format="uri")


class URLPath(pathlib.PosixPath):  # pylint: disable=abstract-method
    """
    URLPath is a nasty hack to ensure that we can defer the downloading of a
    URL passed as a path until later in prediction dispatch.

    It subclasses pathlib.PosixPath only so that it can pass isinstance(_,
    pathlib.Path) checks.
    """

    _path: Optional[Path]

    def __init__(self, *, source: str, filename: str, fileobj: io.IOBase) -> None:  # pylint: disable=super-init-not-called
        if len(filename) > FILENAME_MAX_LENGTH:
            filename = _truncate_filename_bytes(filename, FILENAME_MAX_LENGTH)

        self.source = source
        self.filename = filename
        self.fileobj = fileobj

        self._path = None

    def convert(self) -> Path:
        if self._path is None:
            dest = tempfile.NamedTemporaryFile(suffix=self.filename, delete=False)  # pylint: disable=consider-using-with
            shutil.copyfileobj(self.fileobj, dest)
            self._path = Path(dest.name)
        return self._path

    def unlink(self, missing_ok: bool = False) -> None:
        if self._path:
            self._path.unlink(missing_ok=missing_ok)

    def __str__(self) -> str:
        # FastAPI's jsonable_encoder will encode subclasses of pathlib.Path by
        # calling str() on them
        return self.source


class URLFile(io.IOBase):
    """
    URLFile is a proxy object for a :class:`urllib3.response.HTTPResponse`
    object that is created lazily. It's a file-like object constructed from a
    URL that can survive pickling/unpickling.
    """

    __slots__ = ("__target__", "__url__", "name")

    def __init__(self, url: str, filename: Optional[str] = None) -> None:
        parsed = urllib.parse.urlparse(url)
        if parsed.scheme not in {
            "http",
            "https",
        }:
            raise ValueError(
                "URLFile requires URL to conform to HTTP or HTTPS protocol"
            )
        object.__setattr__(self, "name", os.path.basename(parsed.path))
        object.__setattr__(self, "__url__", url)

        if parsed.scheme not in {
            "http",
            "https",
        }:
            raise ValueError(
                "URLFile requires URL to conform to HTTP or HTTPS protocol"
            )

        if not filename:
            filename = os.path.basename(parsed.path)

        object.__setattr__(self, "name", filename)
        object.__setattr__(self, "__url__", url)

    def __del__(self) -> None:
        try:
            object.__getattribute__(self, "__target__")
        except AttributeError:
            # Do nothing when tearing down the object if the response object
            # hasn't been created yet.
            return

        super().__del__()

    # We provide __getstate__ and __setstate__ explicitly to ensure that the
    # object is always picklable.
    def __getstate__(self) -> Dict[str, Any]:
        return {
            "name": object.__getattribute__(self, "name"),
            "url": object.__getattribute__(self, "__url__"),
        }

    def __setstate__(self, state: Dict[str, Any]) -> None:
        object.__setattr__(self, "name", state["name"])
        object.__setattr__(self, "__url__", state["url"])

    # Proxy getattr/setattr/delattr through to the response object.
    def __setattr__(self, name: str, value: Any) -> None:
        if hasattr(type(self), name):
            object.__setattr__(self, name, value)
        else:
            setattr(self.__wrapped__, name, value)

    def __getattr__(self, name: str) -> Any:
        if name in ("__target__", "__wrapped__", "__url__"):
            raise AttributeError(name)
        elif name == "name":
            return object.__getattribute__(self, "name")
        return getattr(self.__wrapped__, name)

    def __delattr__(self, name: str) -> None:
        if hasattr(type(self), name):
            object.__delattr__(self, name)
        else:
            delattr(self.__wrapped__, name)

    # Luckily the only dunder method on HTTPResponse is __iter__
    def __iter__(self) -> Iterator[bytes]:
        return iter(self.__wrapped__)

    @property
    def __wrapped__(self) -> Any:
        try:
            return object.__getattribute__(self, "__target__")
        except AttributeError:
            pass
        url = object.__getattribute__(self, "__url__")
        resp = requests.get(url, stream=True, timeout=10)
        resp.raise_for_status()
        resp.raw.decode_content = True
        object.__setattr__(self, "__target__", resp.raw)
        return resp.raw

    def __repr__(self) -> str:
        try:
            target = object.__getattribute__(self, "__target__")
        except AttributeError:
            return f"<{type(self).__name__} at 0x{id(self):x} for {object.__getattribute__(self, '__url__')!r}>"

        return f"<{type(self).__name__} at 0x{id(self):x} wrapping {target!r}>"


def get_filename(url: str) -> str:
    parsed_url = urllib.parse.urlparse(url)

    if parsed_url.scheme == "data":
        with urllib.request.urlopen(url) as resp:  # noqa: S310
            mime_type = resp.headers.get_content_type()
            extension = mimetypes.guess_extension(mime_type)
            if extension is None:
                return "file"
            return "file" + extension

    basename = os.path.basename(parsed_url.path)
    basename = urllib.parse.unquote_plus(basename)

    # If the filename is too long, we truncate it (appending '~' to denote the
    # truncation) while preserving the file extension.
    # - truncate it
    # - append a tilde
    # - preserve the file extension
    if _len_bytes(basename) > FILENAME_MAX_LENGTH:
        basename = _truncate_filename_bytes(basename, length=FILENAME_MAX_LENGTH)

    for c in FILENAME_ILLEGAL_CHARS:
        basename = basename.replace(c, "_")

    return basename


Item = TypeVar("Item")
_concatenate_iterator_schema = {
    "type": "array",
    "items": {"type": "string"},
    "x-cog-array-type": "iterator",
    "x-cog-array-display": "concatenate",
}


class ConcatenateIterator(Iterator[Item]):  # pylint: disable=abstract-method
    @classmethod
    def validate(cls, value: Iterator[Any]) -> Iterator[Any]:
        return value

    if PYDANTIC_V2:
        from pydantic import GetCoreSchemaHandler
        from pydantic.json_schema import JsonSchemaValue
        from pydantic_core import CoreSchema

        @classmethod
        def __get_pydantic_core_schema__(
            cls,
            source: Type[Any],  # pylint: disable=unused-argument
            handler: "pydantic.GetCoreSchemaHandler",  # pylint: disable=unused-argument
        ) -> "CoreSchema":
            from pydantic_core import (  # pylint: disable=import-outside-toplevel
                core_schema,
            )

            return core_schema.union_schema(
                [
                    core_schema.is_instance_schema(Iterator),
                    core_schema.no_info_plain_validator_function(cls.validate),
                ]
            )

        @classmethod
        def __get_pydantic_json_schema__(
            cls, core_schema: "CoreSchema", handler: "pydantic.GetJsonSchemaHandler"
        ) -> "JsonSchemaValue":  # type: ignore # noqa: F821
            json_schema = handler(core_schema)
            json_schema.pop("allOf", None)
            json_schema.update(_concatenate_iterator_schema)
            return json_schema

    else:

        @classmethod
        def __get_validators__(cls) -> Iterator[Any]:
            yield cls.validate

        @classmethod
        def __modify_schema__(cls, field_schema: Dict[str, Any]) -> None:
            """Defines what this type should be in openapi.json"""
            field_schema.pop("allOf", None)
            field_schema.update(_concatenate_iterator_schema)


class AsyncConcatenateIterator(AsyncIterator[Item]):
    @classmethod
    def validate(cls, value: AsyncIterator[Any]) -> AsyncIterator[Any]:
        return value

    if PYDANTIC_V2:
        from pydantic import GetCoreSchemaHandler
        from pydantic.json_schema import JsonSchemaValue
        from pydantic_core import CoreSchema

        @classmethod
        def __get_pydantic_core_schema__(
            cls,
            source: Type[Any],  # pylint: disable=unused-argument
            handler: "pydantic.GetCoreSchemaHandler",  # pylint: disable=unused-argument
        ) -> "CoreSchema":
            from pydantic_core import (  # pylint: disable=import-outside-toplevel
                core_schema,
            )

            return core_schema.union_schema(
                [
                    core_schema.is_instance_schema(AsyncIterator),
                    core_schema.no_info_plain_validator_function(cls.validate),
                ]
            )

        @classmethod
        def __get_pydantic_json_schema__(
            cls, core_schema: "CoreSchema", handler: "pydantic.GetJsonSchemaHandler"
        ) -> "JsonSchemaValue":  # type: ignore # noqa: F821
            json_schema = handler(core_schema)
            json_schema.pop("allOf", None)
            json_schema.update(_concatenate_iterator_schema)
            return json_schema
    else:

        @classmethod
        def __modify_schema__(cls, field_schema: Dict[str, Any]) -> None:
            """Defines what this type should be in openapi.json"""
            field_schema.pop("allOf", None)
            field_schema.update(_concatenate_iterator_schema)

        @classmethod
        def __get_validators__(cls) -> Iterator[Any]:
            yield cls.validate


Weights = Union[File, Path, str]


def get_filename_from_urlopen(resp: urllib.response.addinfourl) -> str:
    mime_type = resp.headers.get_content_type()
    extension = mimetypes.guess_extension(mime_type)
    return ("file" + extension) if extension else "file"


def _len_bytes(s: str, encoding: str = "utf-8") -> int:
    return len(s.encode(encoding))


def _truncate_filename_bytes(s: str, length: int, encoding: str = "utf-8") -> str:
    """
    Truncate a filename to at most `length` bytes, preserving file extension
    and avoiding text encoding corruption from truncation.
    """
    root, ext = os.path.splitext(s.encode(encoding))
    ext = ext.decode(encoding).split("?")[0].encode(encoding)
    root = root[: length - len(ext) - 1]
    return root.decode(encoding, "ignore") + "~" + ext.decode(encoding)



from typing import List

from cog import BasePredictor, Path


class Predictor(BasePredictor):
    def predict(self) -> List[Path]:
        predictions = ["foo", "bar", "baz"]
        output = []
        for i, prediction in enumerate(predictions):
            out_path = Path(f"/tmp/out-{i}.txt")
            with out_path.open("w") as f:
                f.write(prediction)
            output.append(out_path)
        return output



from cog import BasePredictor, Path

class Predictor(BasePredictor):
   def predict(self, paths: list[Path]) -> str:
       output_parts = []  # Use a list to collect file contents
       for path in paths:
           with open(path) as f:
             output_parts.append(f.read())
       return "".join(output_parts)
